# tag::enumimport[]
import enum
# end::enumimport[]
# tag::namedtuple[]
from collections import namedtuple
# end::namedtuple[]

# tag::color[]
class Player(enum.Enum):
    black = 1
    white = 2

    @property
    def other(self):
        return Player.black if self == Player.white else Player.white
# end::color[]


# tag::points[]
class Point(namedtuple('Point', 'row col')):
    def neighbors(self):
        return [
            Point(self.row - 1, self.col),
            Point(self.row + 1, self.col),
            Point(self.row, self.col - 1),
            Point(self.row, self.col + 1),
        ]
# end::points[]

    def __deepcopy__(self, memodict={}):
        # These are very immutable.
        return self


from abc import ABC, abstractmethod
from typing import Generic, Set, Tuple, TypeVar

###############################################################################
# An AdversarialSearchProblem is a representation of a game that is convenient
# for running adversarial search algorithms.
#
# A game can be put into this form by extending the AdversarialSearchProblem
# class. See tttproblem.py for an example of this.
#
# Every subclass of AdversarialSearchProblem has its game states represented
# as instances of a subclass of GameState. The only requirement that of a
# subclass of GameState is that it must implement that player_to_move(.) method,
# which returns the index (0-indexed) of the next player to move.
###############################################################################


class GameState(ABC):
    @abstractmethod
    def player_to_move(self) -> int:
        """
        Output- Returns the index of the player who will move next.
        """
        pass


State = TypeVar("State", bound=GameState)

# Action represents the type of actions that an instance of AdversarialSearchProblem uses to
# cause a transition. It's generic because different games have different actions: TTT requires
# placing a piece on a *2D* grid, while Connect 4 just involves selecting a column.
Action = TypeVar("Action")


class AdversarialSearchProblem(ABC, Generic[State, Action]):
    def get_start_state(self):
        """
        Output- Returns the state from which to start.
        """
        return self._start_state

    def set_start_state(self, state: State):
        """
        Changes the start state to the given state.
        Note to student: You should not need to use this.
        This is only for running games.

        Input:
                state- a GameState
        """
        self._start_state = state

    @abstractmethod
    def get_available_actions(self, state: State) -> Set[Action]:
        """
        Input:
                state- a GameState
        Output:
                Returns the set of actions available to the player-to-move
                from the given state
        """
        pass

    @abstractmethod
    def transition(self, state: State, action: Action) -> State:
        """
        Input:
                state- a Gamestate
                action- the action to take
        Ouput:
                Returns the state that results from taking the given action
                from the given state. (Assume deterministic transitions.)
        """
        assert not (self.is_terminal_state(state))
        assert action in self.get_available_actions(state)
        pass

    @abstractmethod
    def is_terminal_state(self, state: State) -> bool:
        """
        Input:
                state: a GameState
        Output:
                Returns a boolean indicating whether or not the given
                state is terminal.
        """
        pass

    # Used to be called evaluate_state
    @abstractmethod
    def evaluate_terminal(self, state: State) -> Tuple[int, int]:
        """
        Should be called when determining which player benefits from a given *terminal* state.
        The range of values returned here should be synchronized with heuristic_func.

        Because we're evaluating terminal states, we're essentially evaluating losing, winning, and
        tieing. You should make sure that the sum of the tuple you return sums to a constant number,
        like 1. If player 0 wins, then should their score be high or low relative to player 1?

        Final note: evaluate_terminal and heuristic_func do very similar things. In fact, their
        ranges are the same! However, we split these up because heuristic_func should be used in
        only the algorithm that uses a heuristic, whereas evaluate_terminal is used across all
        algorithms, since they all need to know how good or bad a terminal state is.

        Input:
                state: a TERMINAL GameState
        Output:
                Returns a Tuple of player 0's value and player 1's value, where each value
                represents whether the player lost, tied, or won.
        """
        assert self.is_terminal_state(state)
        pass


def HeuristicAdversarialSearchProblem(AdversarialSearchProblem):
    @abstractmethod
    def heuristic(self, state: State) -> float:
        """
        Input:
                state: The current game state.
        Output:
                Returns a heuristic evaluation of state (a float)
        """
        pass


###############################################################################
# GameUI is an abstraction that allows you to interact directly with
# an AdversarialSearchProblem (through gamerunner.py). See tttproblem or
# connect4problem for examples.
#
# Utilizing GameUI is NOT necessary for this assignment, although you can use
# it with any ASPs you may decide to create.
###############################################################################

class GameUI(ABC):
    def update_state(self, state: GameState):
        """
        Updates the state currently being rendered.
        """
        self._state = state

    @abstractmethod
    def render(self):
        """
        Renders the GameUI instance's render (presumably this will be called continuously).
        """
        pass

    @abstractmethod
    def get_user_input_action(self):
        """
        Output- Returns an action obtained through the GameUI input itself.
        (It is expected that GameUI validates that the action is valid).
        """
        pass



import platform
import subprocess

import numpy as np

# tag::print_utils[]
from agt_arena.game.dlgo_go import gotypes

COLS = 'ABCDEFGHJKLMNOPQRST'
STONE_TO_CHAR = {
    None: ' . ',
    gotypes.Player.black: ' x ',
    gotypes.Player.white: ' o ',
}


def print_move(player, move):
    if move.is_pass:
        move_str = 'passes'
    elif move.is_resign:
        move_str = 'resigns'
    else:
        move_str = '%s%d' % (COLS[move.point.col - 1], move.point.row)
    print('%s %s' % (player, move_str))


def print_board(board):
    for row in range(board.num_rows, 0, -1):
        bump = " " if row <= 9 else ""
        line = []
        for col in range(1, board.num_cols + 1):
            stone = board.get(gotypes.Point(row=row, col=col))
            line.append(STONE_TO_CHAR[stone])
        print('%s%d %s' % (bump, row, ''.join(line)))
    print('    ' + '  '.join(COLS[:board.num_cols]))
# end::print_utils[]


# tag::human_coordinates[]
def point_from_coords(coords):
    col = COLS.index(coords[0]) + 1
    row = int(coords[1:])
    return gotypes.Point(row=row, col=col)
# end::human_coordinates[]


def coords_from_point(point):
    return '%s%d' % (
        COLS[point.col - 1],
        point.row
    )

def clear_screen():
    # see https://stackoverflow.com/a/23075152/323316
    if platform.system() == "Windows":
        subprocess.Popen("cls", shell=True).communicate()
    else:  # Linux and Mac
        # the link uses print("\033c", end=""), but this is the original sequence given in the book.
        print(chr(27) + "[2J")

# NOTE: MoveAge is only used in chapter 13, and doesn't make it to the main text.
# This feature will only be implemented in goboard_fast.py so as not to confuse
# readers in early chapters.
class MoveAge():
    def __init__(self, board):
        self.move_ages = - np.ones((board.num_rows, board.num_cols))

    def get(self, row, col):
        return self.move_ages[row, col]

    def reset_age(self, point):
        self.move_ages[point.row - 1, point.col - 1] = -1

    def add(self, point):
        self.move_ages[point.row - 1, point.col - 1] = 0

    def increment_all(self):
        self.move_ages[self.move_ages > -1] += 1


import copy
from agt_arena.game.dlgo_go.gotypes import Player, Point
from agt_arena.game.dlgo_go.scoring import compute_game_result
from agt_arena.game.dlgo_go import zobrist
from agt_arena.game.dlgo_go.utils import MoveAge

neighbor_tables = {}
corner_tables = {}


def init_neighbor_table(dim):
    rows, cols = dim
    new_table = {}
    for r in range(1, rows + 1):
        for c in range(1, cols + 1):
            p = Point(row=r, col=c)
            full_neighbors = p.neighbors()
            true_neighbors = [
                n for n in full_neighbors
                if 1 <= n.row <= rows and 1 <= n.col <= cols]
            new_table[p] = true_neighbors
    neighbor_tables[dim] = new_table


def init_corner_table(dim):
    rows, cols = dim
    new_table = {}
    for r in range(1, rows + 1):
        for c in range(1, cols + 1):
            p = Point(row=r, col=c)
            full_corners = [
                Point(row=p.row - 1, col=p.col - 1),
                Point(row=p.row - 1, col=p.col + 1),
                Point(row=p.row + 1, col=p.col - 1),
                Point(row=p.row + 1, col=p.col + 1),
            ]
            true_corners = [
                n for n in full_corners
                if 1 <= n.row <= rows and 1 <= n.col <= cols]
            new_table[p] = true_corners
    corner_tables[dim] = new_table


class IllegalMoveError(Exception):
    pass


class GoString():
    """Stones that are linked by a chain of connected stones of the
    same color.
    """
    def __init__(self, color, stones, liberties):
        self.color = color
        self.stones = frozenset(stones)
        self.liberties = frozenset(liberties)

    def without_liberty(self, point):
        new_liberties = self.liberties - set([point])
        return GoString(self.color, self.stones, new_liberties)

    def with_liberty(self, point):
        new_liberties = self.liberties | set([point])
        return GoString(self.color, self.stones, new_liberties)

    def merged_with(self, string):
        """Return a new string containing all stones in both strings."""
        assert string.color == self.color
        combined_stones = self.stones | string.stones
        return GoString(
            self.color,
            combined_stones,
            (self.liberties | string.liberties) - combined_stones)

    @property
    def num_liberties(self):
        return len(self.liberties)

    def __eq__(self, other):
        return isinstance(other, GoString) and \
            self.color == other.color and \
            self.stones == other.stones and \
            self.liberties == other.liberties

    def __deepcopy__(self, memodict={}):
        return GoString(self.color, self.stones, copy.deepcopy(self.liberties))


class Board():
    def __init__(self, num_rows=19, num_cols=19):
        self.num_rows = num_rows
        self.num_cols = num_cols
        self._grid = {}
        self._hash = zobrist.EMPTY_BOARD
        self.black_captures = 0  # stones captured by black
        self.white_captures = 0  # stones captured by white

        global neighbor_tables
        dim = (num_rows, num_cols)
        if dim not in neighbor_tables:
            init_neighbor_table(dim)
        if dim not in corner_tables:
            init_corner_table(dim)
        self.neighbor_table = neighbor_tables[dim]
        self.corner_table = corner_tables[dim]
        self.move_ages = MoveAge(self)


    def neighbors(self, point):
        return self.neighbor_table[point]

    def corners(self, point):
        return self.corner_table[point]

    def place_stone(self, player, point):
        assert self.is_on_grid(point)
        if self._grid.get(point) is not None:
            print('Illegal play on %s' % str(point))
            return False

        adjacent_same_color = []
        adjacent_opposite_color = []
        liberties = []

        for neighbor in point.neighbors():
            if not self.is_on_grid(neighbor):
                continue
            neighbor_string = self._grid.get(neighbor)
            if neighbor_string is None:
                liberties.append(neighbor)
            elif neighbor_string.color == player:
                adjacent_same_color.append(neighbor_string)
            else:
                adjacent_opposite_color.append(neighbor_string)

        new_string = GoString(player, [point], liberties)

        # Merge adjacent strings of the same color
        for same_color_string in adjacent_same_color:
            new_string = new_string.merged_with(same_color_string)
        for new_string_point in new_string.stones:
            self._grid[new_string_point] = new_string

        # Remove captured stones
        for other_color_string in adjacent_opposite_color:
            if other_color_string.num_liberties == 0:
                self.remove_string(other_color_string)
                # Update capture counts
                if player == Player.black:
                    self.black_captures += len(other_color_string.stones)
                else:
                    self.white_captures += len(other_color_string.stones)

        return True

    def remove_string(self, string):
        for point in string.stones:
            # Remove stones from the board
            del self._grid[point]
            # Update liberties of adjacent strings
            for neighbor in point.neighbors():
                if not self.is_on_grid(neighbor):
                    continue
                neighbor_string = self._grid.get(neighbor)
                if neighbor_string is None:
                    continue
                if neighbor_string is not string:
                    neighbor_string.add_liberty(point)

    @property
    def captured_stones(self):
        """Total number of captured stones"""
        return self.black_captures + self.white_captures

    def get_captures_by_player(self, player):
        """Get number of stones captured by a specific player"""
        return self.black_captures if player == Player.black else self.white_captures

    def is_self_capture(self, player, point):
        friendly_strings = []
        for neighbor in self.neighbor_table[point]:
            neighbor_string = self._grid.get(neighbor)
            if neighbor_string is None:
                # This point has a liberty. Can't be self capture.
                return False
            elif neighbor_string.color == player:
                # Gather for later analysis.
                friendly_strings.append(neighbor_string)
            else:
                if neighbor_string.num_liberties == 1:
                    # This move is real capture, not a self capture.
                    return False
        if all(neighbor.num_liberties == 1 for neighbor in friendly_strings):
            return True
        return False

    def will_capture(self, player, point):
        for neighbor in self.neighbor_table[point]:
            neighbor_string = self._grid.get(neighbor)
            if neighbor_string is None:
                continue
            elif neighbor_string.color == player:
                continue
            else:
                if neighbor_string.num_liberties == 1:
                    # This move would capture.
                    return True
        return False

    def is_on_grid(self, point):
        return 1 <= point.row <= self.num_rows and \
            1 <= point.col <= self.num_cols

    def get(self, point):
        """Return the content of a point on the board.

        Returns None if the point is empty, or a Player if there is a
        stone on that point.
        """
        string = self._grid.get(point)
        if string is None:
            return None
        return string.color

    def get_go_string(self, point):
        """Return the entire string of stones at a point.

        Returns None if the point is empty, or a GoString if there is
        a stone on that point.
        """
        string = self._grid.get(point)
        if string is None:
            return None
        return string

    def __eq__(self, other):
        return isinstance(other, Board) and \
            self.num_rows == other.num_rows and \
            self.num_cols == other.num_cols and \
            self._hash() == other._hash()

    def __deepcopy__(self, memodict={}):
        copied = Board(self.num_rows, self.num_cols)
        # Can do a shallow copy b/c the dictionary maps tuples
        # (immutable) to GoStrings (also immutable)
        copied._grid = copy.copy(self._grid)
        copied._hash = self._hash
        return copied

# tag::return_zobrist[]
    def zobrist_hash(self):
        return self._hash
# end::return_zobrist[]


class Move():
    """Any action a player can play on a turn.

    Exactly one of is_play, is_pass, is_resign will be set.
    """
    def __init__(self, point=None, is_pass=False, is_resign=False):
        assert (point is not None) ^ is_pass ^ is_resign
        self.point = point
        self.is_play = (self.point is not None)
        self.is_pass = is_pass
        self.is_resign = is_resign

    @classmethod
    def play(cls, point):
        """A move that places a stone on the board."""
        return Move(point=point)

    @classmethod
    def pass_turn(cls):
        return Move(is_pass=True)

    @classmethod
    def resign(cls):
        return Move(is_resign=True)

    def __str__(self):
        if self.is_pass:
            return 'pass'
        if self.is_resign:
            return 'resign'
        return '(r %d, c %d)' % (self.point.row, self.point.col)

    def __hash__(self):
        return hash((
            self.is_play,
            self.is_pass,
            self.is_resign,
            self.point))

    def  __eq__(self, other):
        return (
            self.is_play,
            self.is_pass,
            self.is_resign,
            self.point) == (
            other.is_play,
            other.is_pass,
            other.is_resign,
            other.point)


class GameState():
    def __init__(self, board, next_player, previous, move):
        self.board = board
        self.next_player = next_player
        self.previous_state = previous
        if previous is None:
            self.previous_states = frozenset()
        else:
            self.previous_states = frozenset(
                previous.previous_states |
                {(previous.next_player, previous.board.zobrist_hash())})
        self.last_move = move

    def apply_move(self, move):
        """Return the new GameState after applying the move."""
        if move.is_play:
            next_board = copy.deepcopy(self.board)
            next_board.place_stone(self.next_player, move.point)
        else:
            next_board = self.board
        return GameState(next_board, self.next_player.other, self, move)

    @classmethod
    def new_game(cls, board_size):
        if isinstance(board_size, int):
            board_size = (board_size, board_size)
        board = Board(*board_size)
        return GameState(board, Player.black, None, None)

    def is_move_self_capture(self, player, move):
        if not move.is_play:
            return False
        return self.board.is_self_capture(player, move.point)

    @property
    def situation(self):
        return (self.next_player, self.board)

    def does_move_violate_ko(self, player, move):
        if not move.is_play:
            return False
        if not self.board.will_capture(player, move.point):
            return False
        next_board = copy.deepcopy(self.board)
        next_board.place_stone(player, move.point)
        next_situation = (player.other, next_board.zobrist_hash())
        return next_situation in self.previous_states

    def is_valid_move(self, move):
        if self.is_over():
            return False
        if move.is_pass or move.is_resign:
            return True
        return (
            self.board.get(move.point) is None and
            not self.is_move_self_capture(self.next_player, move) and
            not self.does_move_violate_ko(self.next_player, move))

    def is_over(self):
        if self.last_move is None:
            return False
        if self.last_move.is_resign:
            return True
        second_last_move = self.previous_state.last_move
        if second_last_move is None:
            return False
        return self.last_move.is_pass and second_last_move.is_pass

    def legal_moves(self):
        moves = []
        for row in range(1, self.board.num_rows + 1):
            for col in range(1, self.board.num_cols + 1):
                move = Move.play(Point(row, col))
                if self.is_valid_move(move):
                    moves.append(move)
        # These two moves are always legal.
        moves.append(Move.pass_turn())
        moves.append(Move.resign())

        return moves

    def winner(self):
        if not self.is_over():
            return None
        if self.last_move.is_resign:
            return self.next_player
        game_result = compute_game_result(self)
        return game_result.winner


from abc import ABC, abstractmethod

class BaseMatchmaker(ABC):
    """
    Abstract base class for all matchmakers.
    """
    
    def __init__(self):
        self.bot_data = {}

    def update_bot_data(self, bot, data):
        """Update data for a bot (e.g., ELO)"""
        if bot not in self.bot_data:
            raise ValueError(f"Bot {bot} is not registered.")
        self.bot_data[bot].update(data)

    @abstractmethod
    def generate_matches(self, players):
        """
        Generate a list of matches from the given players.
        Args:
            players: List of players (e.g., ["Player 1", "Player 2", ...]).
        Returns:
            List of matches, where each match is a tuple of players.
        """
        pass


import numpy as np
import pyspiel
import sys


def create_go_game(size):
    """
    load open-spiel game with provided size
    """
    if size == 5:
        komi = 0.5
    elif size == 9:
        komi = 5.5
    else:
        komi = 7.5
    game = pyspiel.load_game("go", {"board_size": size, "komi": komi})
    state = game.new_initial_state()
    return state



from typing import Sequence, Type
import json
import numpy as np
from agt_arena.game.go.adversarial_search_problem import AdversarialSearchProblem, GameState
import copy
from  agt_arena.game.go.go_utils import create_go_game
from pyspiel import Game

Action = int

DEFAULT_SIZE = 9

class GoState(GameState):
    """
    A state of the game of Go.
    Includes methods and properties for the state of the board, player to move, and other useful methods
    """
    def __init__(self, pyspiel_state: Game, player_to_move: int = 0):
        """
        Initialize GoState with pyspiel as backend Go engine.
        The initial state is created with a call to create_go_game() in go_utils.py
        Every other state will be generated from applying actions to the initial state.
        This essentially functions as a wrapper class to conver pyspiel game states to 
        The ASP interface used previously.

        :param pyspiel_state: pyspiel state of the game
        :param player_to_move: player to move
        """
        self.internal_state = pyspiel_state
        self.size = int(np.sqrt(len(pyspiel_state.observation_tensor()) / 4))


    def player_to_move(self) -> int:
        """
        Get the current player to move
        :return: player to move BLACK (0) or WHITE (1)
        """
        return self.internal_state.current_player()

    def get_board(self) -> np.ndarray:
        """
        Return the current board as a numpy array
        The board will have shape (4, size, size)
        The first channel (i.e., get_board()[0]) is the board for BLACK. There are 1's where the black pieces are and 0's elsewhere.
        The second channel (i.e., get_board()[1]) is the board for WHITE. There are 1's where the white pieces are and 0's elsewhere.
        The third channel (i.e., get_board()[2]) is the board for EMPTY. There are 1's where the empty spaces are and 0's elsewhere.
        The fourth channel (i.e., get_board()[3]) is the board for whose turn it is. There are 0's when it is BLACK's turn and 1's when it is white's.

        This is the default observation tensor used by pyspiel.
        """
        return np.array(self.internal_state.observation_tensor(0)).reshape(-1, self.size, self.size)

    def terminal_value(self) -> float:
        """
        Return the terminal value of the game.
        :return: 1 if BLACK wins, -1 if WHITE wins
        """
        return self.internal_state.returns()

    def clone(self) -> GameState:
        """
        Create a copy of the current game state.
        This is used for safety with the game runner.
        We don't want search algorithms to be able to directly modify the game state,
        so we only pass a copy of the state to the search algorithms.
        :return: a copy of the current game state
        """
        return GoState(self.internal_state.clone(), self.internal_state.current_player())

    def is_terminal_state(self) -> bool:
        """
        Checks if the game is in a terminal state.
        The state is if there are no legal actions left or the players have passed twice in a row.

        :return: True if the game is in a terminal state, False otherwise
        """
        return self.internal_state.is_terminal()

    def legal_actions(self) -> Sequence[Action]:
        """
        Return all possible legal actions for the given state.
        Note: Actions are represented as integers, by default.
        For a more human-readable representation, use action_index_to_coord()

        NOTE: It is preferrable to get the available actions from the search problem,
        not this state. 

        :return: list of legal actions
        """
        return self.internal_state.legal_actions()

    def apply_action(self, action: Action):
        """
        Apply action and update internal state.
        Action must be an int, not a coordinate.

        NOTE: It is preferrable to use the transition function from the search problem,
        not this method to apply actions. 
        """
        self.internal_state.apply_action(action)

    def get_pieces_coordinates(self, player_index: int):
        """
        Get the indices of the pieces of the given player.
        :param player_index: 0 for BLACK, 1 for WHITE
        :return: list of coordinates of the pieces of the given player
        """
        player_board = np.array(self.internal_state.observation_tensor(
            0)).reshape((-1, self.size, self.size))[player_index]
        return np.argwhere(player_board == 1)

    def get_pieces_array(self, player_index):
        """
        Get the 2D array of the pieces of the given player.
        The array will have shape (size, size) and will have 1's where the pieces are and 0's elsewhere.

        :param player_index: 0 for BLACK, 1 for WHITE
        :return: 2D np array of the pieces of the given player
        """
        player_board = np.array(self.internal_state.observation_tensor(
            0)).reshape((-1, self.size, self.size))[player_index]
        return player_board

    def get_empty_spaces(self):
        """
        return a 2D array of the empty spaces on the board
        The array will have shape (size, size) and will have 1's where the empty spaces are and 0's elsewhere.

        :return: 2D np array of the empty spaces on the board
        """
        return self.internal_state.observation_tensor(2)
    
    def action_index_to_coord(self, action: Action) -> tuple[int, int]:
        """
        Convert an action index to a coordinate.
        :param action: action index
        :return: coordinate (x, y)
        """
        return (action % self.size, action // self.size)

    def serialize(self) -> str:
        """
        Serialize the current GoState into a JSON string using pyspiel's serialize method.
        Includes the pyspiel state and the player to move.
        :return: JSON string representing the game state
        """
        return json.dumps({
            "pyspiel_state": self.internal_state.serialize(),
            "player_to_move": self.player_to_move(),
        })

    @staticmethod
    def deserialize(serialized_str: str, pyspiel_game: Game) -> 'GoState':
        """
        Deserialize a JSON string to recreate a GoState object using pyspiel's deserialize_state method.
        :param serialized_str: JSON string representing the game state
        :param pyspiel_game: pyspiel Game object for the game of Go
        :return: GoState object
        """
        data = json.loads(serialized_str)
        pyspiel_state = pyspiel_game.deserialize_state(data["pyspiel_state"])
        return GoState(pyspiel_state, data["player_to_move"])

    def __repr__(self):
        return str(self.internal_state)


class GoProblem(AdversarialSearchProblem[GoState, Action]):
    def __init__(self, size=DEFAULT_SIZE, state=None, player_to_move=0):
        """
        Create a new Go search problem.
        If no state is provided, a new game is created with the given size.
        """
        if state is None:
            game_state = create_go_game(size)
        else:
            game_state = state
        self.start_state = GoState(game_state, player_to_move)

    def get_available_actions(self, state: GoState) -> Sequence[Action]:
        """
        Get the available actions for the given state.
        Use this to get the list of available actions for a given state.
        Note: An action in this case is an integer in range [0, size^2].
        Each action index corresponds to a coordinate on the board (x, y) = (action % size, action // size).
        With action=size**2 reserved for the pass action.

        :param state: current state
        :return: list of available actions
        """
        return state.legal_actions()

    def transition(self, state: GoState, action: Action) -> GoState:
        """
        Return new_state resulting from applying action to state.

        :param state: current state
        :param action: action to apply
        :return: new state resulting from applying action to state
        """
        new_state = state.clone()
        new_state.apply_action(action)
        return new_state

    def is_terminal_state(self, state: GoState) -> bool:
        """
        Return if the given state is a terminal state.
        State is terminal if no legal actions are available or the players have passed twice in a row.

        :param state: current state
        :return: True if the state is terminal, False otherwise
        """
        return state.is_terminal_state()

    def evaluate_terminal(self, state: GoState) -> float:
        """
        Get the value of the terminal state.
        The value is 1 if BLACK wins and -1 if WHITE wins.

        :param state: current state
        :return: value of the terminal state
        """
        return state.terminal_value()[0]

    def action_index_to_string(self, action: Action) -> str:
        """
        Convert an Action (index) to a string.
        """
        return "(" + str(action % self.start_state.size) + ", " +  str(action // self.start_state.size) + ")"


from flask import Flask, request, jsonify
import os
import pyspiel
from agt_arena.game.go.go_search_problem import GoState


def create_bot_server(bot_factory, board_size):
    """Factory function to create a Flask server for any bot"""
    app = Flask(__name__)
    bot = bot_factory(board_size)
    print(f"BOT IN BOT SERVER AHHHHHHH {bot}")
    game = pyspiel.load_game("go", {"board_size": board_size})

    @app.route('/get_move', methods=['POST'])
    def get_move():
        serialized_state = request.json['game_state']
        time_remaining = request.json['time_remaining']
        game_state = GoState.deserialize(serialized_state, game)
        move = bot.get_move(game_state, time_remaining)
        return jsonify({'move': move})
        
    @app.route('/setup', methods=['POST'])
    def setup():
        player_data = request.json
        bot.on_start(player_data)
        return jsonify({'status': 'ready'})

    @app.route('/update', methods=['POST'])
    def update():
        serialized_state = request.json['game_state']
        game_state = GoState.deserialize(serialized_state, game)
        bot.on_update(game_state)
        return jsonify({'status': 'updated'})

    @app.route('/health', methods=['GET'])
    def health_check():
        print(str(bot))
        if str(bot) is None:
            return jsonify({'status': 'unhealthy'}), 100
        else:
            return jsonify({'status': 'healthy', 'name': str(bot)}), 200

    return app


def run_bot_server(bot_factory, board_size, port):
    """Convenience function to create and run a bot server"""
    print(f"Running bot server for {bot_factory.__name__} with board size {board_size}")
    app = create_bot_server(bot_factory, board_size)
    app.run(host='0.0.0.0', port=port, debug=True)